/* MiniPOV firmware.
 * (c) Matt Evans, 1 May 2009
 * Version 1.0
 *
 * Bugs: There's an odd blip at the end of displayed sentences.  That
 * is, there's an obvious off-by-one bug for the reader to find.
 */

#include <inttypes.h>
#include "minipov.h"
// Define CPU frequency before including delay.h
// This selects the right IO defs from the -mmcu=atmega48 GCC param.
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h> // CPU freq. defined!
#include <avr/sleep.h>

PROGMEM const char stringA[]  = "Blah blah ";
PROGMEM const char stringB[]  = "Firmware is great! ";
PROGMEM const char stringC[]  = "\136\140\136\140";

PROGMEM const char * const strings[] = { 
	stringA,
	stringB,
	stringC,
};

// This var is number of strings + 1 (the last being ADC readout)
const uint8_t displayViews = (sizeof(strings) / sizeof(char *)) + 1;

uint8_t displayView = 0; // Different phrases
uint8_t ledsVal     = 0;
uint16_t globalTime = 0;
uint16_t blinks     = 0;

uint8_t buttonPressed = 0;

PROGMEM const uint8_t font[] = {
#include "font.inc"
};

uint8_t stringPos = 0;
uint8_t charPos = 0;

void setLEDs(uint8_t d)
{
	uint8_t pa = 0;
	uint8_t pb = 0;

	PORTA |= LEDSA;
	PORTB |= LEDSB;

	pa = ((d & 0x01) ? LED0 : 0) |
		((d & 0x02) ? LED1 : 0) |
		((d & 0x10) ? LED4 : 0) |
		((d & 0x20) ? LED5 : 0) |
		((d & 0x40) ? LED6 : 0) |
		((d & 0x80) ? LED7 : 0);
	pb = ((d & 0x04) ? LED2 : 0) | ((d & 0x08) ? LED3 : 0);

	PORTA &= ~pa;
	PORTB &= ~pb;
}

void ledsOff(void)
{
	PORTA |= LEDSA;
	PORTB |= LEDSB;
}

// Set up pin states
void adcInit(void)
{
	DDRA  &= ~(ADC_IN); // in
	DDRA  |= POT_EN;    // out
	PORTA &= ~(ADC_IN); // Disables pullup
	PORTA |= POT_EN;    // 1 = no current through potentiometer
  
	// ADMUX  (ADC0 = input)
	ADMUX = 0; // ADC0 selected & VCC ref
  
	// Disable digital input on ADC0:
	DIDR0 |= (1<<ADC0D);

	// By default, PRADC=0
	PRR &= ~(1<<PRADC);

	ADCSRA = (1<<ADPS2) | (1<<ADPS1) /*| (1<<ADPS0)*/ | (1<<ADIE);
	ADCSRB = (1<<ADLAR);
}

// Switch ADC on & do a reading when ready (catch IRQ)
void adcOn(void)
{
	// Drag end of pot low
	PORTA &= ~POT_EN;

	// Set ADEN in ADCSRA  (uses power, switch off when we sleep...)
	ADCSRA |= (1<<ADEN);
}

void adcStart(void)
{
	// Start the first conversion, which will begin IRQ delivery:
	ADCSRA |= (1<<ADSC);
}

void adcOff(void)
{
	ADCSRA &= ~(1<<ADEN);
	PORTA |= POT_EN;    // 1 = no current through potentiometer
}

// Set up basic I/O
static void init(void)
{
	PORTA |= LEDSA;
	PORTB |= LEDSB;
	// LED outs
	DDRA |= LEDSA;
	DDRB |= LEDSB;
	// All LEDs off.

	// button in, no pullup. (Shit. solder on a pulldown. aghhh.)
	PORTB &= ~BUTTON;
	DDRB &= ~BUTTON;

	// Button is pinchange irq 1, source 8
	PCMSK1 |= (1<<PCINT8);
	GIMSK |= (1<<PCIE1); // Enable pinchange irq 1
}

void displayTimerInit(void)
{
	OCR0A = 125;     // 125*8=1000, ie 1ms
	TCNT0 = 0;
	TCCR0A = (1<<WGM01);                          // No PWM, mode 2 (CTC)
	TCCR0B = (1<<CS01);  // fclk/8
	TIMSK0 = (1<<OCIE0A);                         // IRQ at match/overflow
}

void displayTimerOff(void)
{
	TIMSK0 = 0;
}

void buttonTimerInit(void)
{
	OCR1A = 1250;    // 1250*8=10000 ie 10ms
	TCNT1 = 0;
	TCCR1A = 0;
	TCCR1B = (1<<WGM12) | (1<<CS11);
	TCCR1C = 0;
	TIMSK1 = (1<<OCIE1A);
}

void buttonTimerOff(void)
{
	TIMSK1 = 0;
}

void buttonHeld(void)
{
	setLEDs(0);
	// Wait for button release!
	while ((PINB & BUTTON) != 0)
	{
	}
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
	adcOff();
	displayTimerOff();
	buttonTimerOff();
	sei(); // ARGH!  We're in an ISR, IRQs are disabled!
	// There's a race there.  Disabled timers so nothing should get a look in...
	sleep_mode();
	cli();
	set_sleep_mode(SLEEP_MODE_IDLE);
	// Wait for button release!
	while ((PINB & BUTTON) != 0)
	{
	}

	displayTimerInit();
	buttonTimerInit();
}

void buttonClick(void)
{
	displayView++;
	if (displayView >= displayViews)
	{
		displayView = 0;

		stringPos = 0;
		charPos = 0;
	}
	// Save view in EEPROM?
}

ISR(PCINT1_vect)
{
	// This does get called!
}

// Sampled buttonness
ISR(TIM1_COMPA_vect)
{
	globalTime++;

	if ((PINB & BUTTON) == 0) // Not pressed
	{
		if (buttonPressed > 10)
		{
			// Released, and has been pressed for a bit
			buttonClick();
		}

		buttonPressed = 0;
	}
	else
	{
		// Pressed
		if (buttonPressed != 0xff)
		{
			buttonPressed++;
		}

		if (buttonPressed > 70)
		{      
			buttonHeld();
			buttonPressed = 0;
		}
	}

	// Kick off a periodic ADC thingy

	if ((globalTime & 0x3fUL) == 0)
	{
		adcOn();
	}
	else if ((globalTime & 0x3fUL) == 1)
	{
		adcStart();
	}
}

// 1KHz tick
ISR(TIM0_COMPA_vect)
{
	blinks++;

	if (displayView == (displayViews - 1))
	{
		if ((blinks & 7) == 0)
		{
			setLEDs(ledsVal);
		}
		else
		{
			setLEDs(0);
		}
	}
	else
	{
		uint8_t slice, ch;
		setLEDs(0);
    
		char *p = pgm_read_word(&strings[displayView]);
		ch = pgm_read_byte(&p[stringPos]);

		if (ch == 0)
		{
			stringPos = 0;
		}
		else
		{
			slice = pgm_read_byte(&font[((ch - 32) * 8) + charPos++]);

			setLEDs(slice);

			if (charPos > 7)
			{
				charPos = 0;
				stringPos++;
			}
		}
	}  
}


ISR(ADC_vect)
{
	uint8_t timePeriod;
	// We got an ADC result!  Read it...
	timePeriod = ledsVal = ADCH;

	if (timePeriod < 5)
	{
		timePeriod = 5;
	}
	OCR0A = timePeriod;

	// One-shot, we switch off & go back to sleep for a few hundred milliseconds to save energy.
	adcOff();
}


int main(void)
{
	init();

	adcInit();

	displayTimerInit();
	buttonTimerInit();
  
	sei();

	set_sleep_mode(SLEEP_MODE_IDLE);

	while(1)
	{
		sleep_mode();
	}
}
